/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

/************************************************************************
*                                                                       *
* 5-26-94 - David Relson                                                *
*                                                                       *
*   When loading this module and the screen height is less than         *
*   33 lines, the colors window displays 128 colors.  When there        *
*   are 33 or more lines, 256 colors are displayed.                     *
*                                                                       *
*   Also, with 33 or more lines the help text is formatted differently. *
*                                                                       *
*   For all screen heights, the color values which used to be simple    *
*   decimal numbers, for example 240, are indicated in hex and as text  *
*   phrases, i.e. color 0xF0 - colBLACK on colINTENSE+colWHITE          *
*                                                                       *
*   A new option W replaces the current level with code to generate     *
*   the current color scheme, i.e. if editting MYCOLORS.E with the      *
*   cursor within the when-loading function, command set-colors option  *
*   W will replace the body of the function with EEL code that will     *
*   generate the current color scheme.                                  *
*                                                                       *
************************************************************************/

#include "eel.h"

#define MAXCOLORS	100	/* max possible # of color classes defined */

#define FG(c)		((c) & 15)	/* extract foreground & background */
#define BG(c)		(((c) >> 4) & mk)
#define COLOR_COL	56
#define COLOR_LINE	2
#define SCRPLACE(c)	(COLOR_COL + FG(c)), (COLOR_LINE + BG(c))

#define	SPACE		' '
#define	COMMA		','

char colorName[][24] = {
    "colBLACK",		// 0
    "colBLUE",		// 1
    "colGREEN",		// 2
    "colCYAN",		// 3
    "colRED",		// 4
    "colMAGENTA",	// 5
    "colBROWN",		// 6
    "colWHITE",		// 7
    "colINTENSE+colBLACK",	// 8+0
    "colINTENSE+colBLUE",	// 8+1
    "colINTENSE+colGREEN",	// 8+2
    "colINTENSE+colCYAN",	// 8+3
    "colINTENSE+colRED",	// 8+4
    "colINTENSE+colMAGENTA",	// 8+5
    "colYELLOW",		// 8+6	(colBROWN+colINTENSE)
    "colINTENSE+colWHITE",      // 8+7
    "8+8INTENSE+colWHITE"	// 8+8
};

// m2u()
//	Convert minus sign to underscore.

char *m2u(p)
char *p;
{
    char *t;
    for (t = p; *t != '\0'; t += 1)
    {
	if (*t == '-')
	    *t = '_';
    }
    return p;
}

char temp[50];

char *fore_back(c)
{
	strcpy(temp, colorName[c & 0x0F]);
	strcat(temp, " on ");
	strcat(temp, colorName[c >> 4]);
	return (temp);
}


char *pad(buf, str, chr, wid)
char *buf, *str;
{
	int l = strlen(str);
	strcpy(buf, str);
	buf[l++] = chr;
	while (l < wid)
		buf[l++] = ' ';
	buf[l] = '\0';
	return (buf);
}


#define LEFTBRACE	"{"
#define RIGHTBRACE	"}"

write_colors(cur_buf)
{
    int i,v;
    char buf0[32], buf1[32], buf2[32];
    int classes = number_of_color_classes();

    bufnum = cur_buf;
    if (re_search(-1, LEFTBRACE))
    {
	int start = point;
	re_search(1, RIGHTBRACE);
	do_save_kill(start, point);
    }

    mark = point;
    bprintf("{\n" );
    bprintf("#define X(num, name, fore, back) alter_color(color_class name, COLOR( fore, back ))\n" );
    for (i = 0; i < classes; i++)
    {
	//		say("color class %s uses color %d.",
	//		    name_color_class(i), alter_color(i, -1));

	v = alter_color(i, -1);
	bprintf("X(0x%02x, %s %s %s);\n",
		v,
		pad(buf0, m2u(name_color_class(i)), COMMA, 20),
		pad(buf1, colorName[v & 0x0F], COMMA, 20),
		pad(buf2, colorName[v >> 4],   SPACE, 20));
    }
    bprintf("}\n" );
    indent_region();
    move_level(-1, RIGHTBRACE, LEFTBRACE);
//  backward_level();
}


when_loading()		/* query user to set colors */
{
    int ht;		// standard window height (10 lines of 25 line screen)
    int mc;		// max colors (128 if 10 lines, 256 if 18)
    int mk = 7;		// mask

    int class, color, i, j, orig[MAXCOLORS], classes;
    int swin, ccwin, cwin, infowin, newcolor, newscheme, orig_scheme;
    char buf[100];
    int sbuf, ccbuf, cbuf, infobuf, orig_win = window_handle, win, page;
    char *new, *old;
    int cur_buf = bufnum;
    jmp_buf this_level;

    save_var top_level = &this_level;
    classes = number_of_color_classes();
    orig_scheme = color_scheme;
    class = iter = page = 0;
    for (i = 0; i < classes; i++)
	orig[i] = alter_color(i, -1);
    sbuf = tmp_buf();	/* list of color schemes */
    ccbuf = tmp_buf();	/* list of color classes */
    cbuf = tmp_buf();	/* window with colors */ 
    infobuf = tmp_buf();	/* help buffer with options */ 

    bufnum = sbuf;	/* now highlight current line in buffers */
    add_region(point_spot, point_spot, color_class menu_highlight,
	       REGLINE);
    bufnum = ccbuf;
    add_region(point_spot, point_spot, color_class menu_highlight,
	       REGLINE);
    for (i = 0; i < classes; i++)
	bprintf("%s\n", name_color_class(i));
    delete(point - 1, point);

winagain:

    if (screen_lines < 18 + 15)
    {
	ht = 10;	// height
	mc = 128;	// max colors
	mk = 7;		// mask
    }
    else
    {
	ht = 18;	// height
	mc = 256;	// max colors
	mk = 15;	// mask
    }

    swin = add_popup(4, 1, 20, ht /* 10 */, ALL_BORD(BSINGLE), sbuf);
    text_color = color_class menu;
    window_title(swin, BTOP, TITLECENTER, " Color Schemes ");
    display_more_msg(swin);

    ccwin = add_popup(29, 1, 21, ht /* 10 */, ALL_BORD(BSINGLE), ccbuf);
    display_column = 0;		/* prevent continuation lines */
    text_color = color_class menu;
    window_title(ccwin, BTOP, TITLECENTER, " Color Classes ");
    display_more_msg(ccwin);

    cwin = add_popup(55, 1, 18, ht /* 10 */, ALL_BORD(BSINGLE), cbuf);
    window_title(cwin, BTOP, TITLECENTER, " Colors ");
    invisible_window = 1;
    insert('\n');
    if (selectable_colors > mc)
	window_title(cwin, BBOTTOM, TITLERIGHT(1), " More ");

    infowin = add_popup(2, ht+4 /* 14 */, 75, ht - 2 /* 8 */, ALL_BORD(BSINGLE), infobuf);
    display_column = 0;		/* prevent continuation lines */
    text_color = color_class menu;
    newscheme = 1;
    display_scroll_bar = 0;
    if (setjmp(top_level))	/* come here if subcommand's aborted */
	;
    for (;;) {
	color = alter_color(class, -1);
	if ((color >= mc) != page)
	    newscheme = 1;
	page = (color >= mc);
	if (newscheme) {
	    window_handle = cwin;
	    point = page;
	    window_handle = swin;
	    buf_zap(sbuf);
	    for (i = 1; i <= final_index(); i++)
		if (name_type(i) == NT_COLSCHEME)
		    bprintf("%s\n", name_name(i));
	    delete(size() - 1, size());	/* no final \n */
	    point = 0;
	    col_search(name_name(color_scheme), 0);
	    j = page ? mc : 0;
	    for (i = 0; i < mc; i++, j++)
		term_write(SCRPLACE(i), "+", 1, j, 0);
	}
	window_handle = ccwin;
	point = 0;
	col_search(name_color_class(class), 0);
	buf_zap(infobuf);
	window_handle = infowin;
	bprintf("Type F or B to change the color scheme, D to "
		"define a new color scheme.\n");
	bprintf("Or use the arrow keys to change the "
		"%s color class.\n", name_color_class(class));
	if (class == color_class text)
	    bprintf("This color is used for normal text.\n");
	else if (class == color_class highlight)
	    bprintf("This color is used for highlighted text.\n");
	else if (class == color_class horiz_border)
	    bprintf("This color is for the line part of each "
		    "mode line between tiled windows.\n");
	else if (class == color_class vert_border)
	    bprintf("This color is for vertical lines "
		    "between tiled windows.\n");
	else if (class == color_class after_exiting)
	    bprintf("Epsilon tries to leave the screen in "
		    "this color when you exit.\n");
	else if (class == color_class debug_text)
	    bprintf("The EEL debugger uses this color.\n");
	else if (class == color_class default)
	    bprintf("Newly-defined color classes use "
		    "this color.\n");
	else if (class == color_class overlap_even)
	    bprintf("Epsilon uses this color when an even "
		    "number of color classes overlap.\n");
	else if (class == color_class overlap_odd)
	    bprintf("Epsilon uses this color when an odd "
		    "number of color classes overlap.\n");
	else
	    bprintf("...testing...\n");
	if (ht < 15)
	{
		bprintf("Type N or P to set other color classes, "
			"T to try out these colors temporarily,\n");
		bprintf("S to set these colors permanently,\n");
		bprintf("W to write the color scheme to buffer 'color-buffer'\n");
		bprintf("or C to cancel.  Color was initially "
			"0x%02x, this color is 0x%02x.", orig[class], color);
	}
	else
	{
		bprintf("Type N or P to set other color classes,\n"
			"T to try out these colors temporarily,\n");
		bprintf("S to set these colors permanently,\n");
		bprintf("W to write the color scheme to buffer '-color-buffer', or\n");
		bprintf("C to cancel.\n");
		bprintf("\n");
		bprintf("Color was initially 0x%02x - %s\n", orig[class], fore_back(orig[class]));
		bprintf("      this color is 0x%02x - %s.",  color, fore_back(color));
	}
	point = 0;
	window_handle = cwin;
	do {
	    maybe_refresh();
	    term_write(SCRPLACE(color), "C", 1, color, 0);
	    full_getkey("Your choice: ", 0);
	    term_write(SCRPLACE(color), "+", 1, color, 0);
	} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
	key = toupper(key);
	newcolor = -100;
	newscheme = 0;
	if (key == 'C' || key == 'Q' || user_abort) {
	    color_scheme = orig_scheme;
	    for (i = 0; i < classes; i++)
		alter_color(i, orig[i]);
	    say("");
	    break;
	} else if (key == 'S' || key == ' ') {
	    show_binding("To remember these colors for "
			 "future sessions, use %s.", "write-state");
	    break;
	} else if (key == 'W') {
	    write_colors(cur_buf);
	    break;
	} else if (IS_MOUSE_DOWN(key)) {
	    win = window_at_coords(mouse_y, mouse_x);
	    if (!is_window(win))
		continue;
	    window_handle = win;
	    get_text_area(win);
	    if (win == cwin) {
		if (bar_on_border()) {
		    i = scroll_bar_code();
		    if ((i ? (i > 0) : !page) != point)
			newcolor = color ^ mc;
		} else {
		    newcolor = (mouse_y - COLOR_LINE) * 16;
		    newcolor += (mouse_x - COLOR_COL);
		    if (page)
			newcolor += mc;
		}
	    } else if (in_text_area(mouse_x, mouse_y))
		point_to_pos(mouse_x, mouse_y);
	    else if (display_scroll_bar && bar_on_border())
		mouse_scroll_bar();
	    else
		continue;
	    if (win == ccwin)
		class = lines_between(0, point);
	    else if (win == swin)
		newscheme = 1;
	    else if (win == infowin && isupper(curchar())
		     && character(point - 1) == ' '
		     && character(point + 1) == ' ')
		ungot_key = curchar();
	}
	switch (get_direction()) {
	    case BTOP:	newcolor = color + 256 - 16; break;
	    case BBOTTOM:	newcolor = color + 16; break;
	    case BLEFT:	newcolor = color + 256 - 1; break;
	    case BRIGHT:	newcolor = color + 1; break;
	}
	switch (key) {
	    case 'N': class++; break;
	    case 'P': class--; break;
	    case 'T':
		show_binding("Continue changing colors "
			     "by typing %s", "exit-level");
		remove_window(swin);
		remove_window(ccwin);
		remove_window(cwin);
		remove_window(infowin);
		recursive_edit();
		goto winagain;
		break;
	    case 'F':
		window_handle = swin;
		if (!nl_forward())
		    point = 0;
		newscheme = 1;
		break;
	    case 'B':
		window_handle = swin;
		if (!nl_reverse())
		    point = size();
		newscheme = 1;
		break;
	    case 'D':
		maybe_refresh();
		get_string(buf, "New scheme name: ");
		if (*buf) {
		    new_variable(buf, NT_COLSCHEME, 0, 0);
		    new = varptr(find_index(buf));
		    old = varptr(color_scheme);
		    for (i = 0; i < classes; i++)
			*new++ = *old++;
		    newscheme = 1;
		}
		break;
	    default:
		if (root_keys[key] == (short) help
		    || key == 'h' || key == '?')
		    help_on_command("set-color");
		else if (IS_TRUE_KEY(key)
			 && newcolor == -100) {
		    maybe_ding(bell_on_bad_key);
		    newscheme = 1;
		}
		break;
	}
	if (newcolor != -100)	/* did it change? */
	    alter_color(class, newcolor % selectable_colors);
	else if (newscheme) {
	    window_handle = swin;
	    grab_full_line(bufnum, buf);
	    i = find_index(buf);
	    if (i > 0 && name_type(i) == NT_COLSCHEME)
		color_scheme = i;
	} else
	    class = (class + classes) % classes;
    }
    remove_window(swin);
    remove_window(ccwin);
    remove_window(cwin);
    remove_window(infowin);
    if (is_window(orig_win))
	window_handle = orig_win;
    buf_delete(sbuf);
    buf_delete(ccbuf);
    buf_delete(cbuf);
    buf_delete(infobuf);
    full_redraw = 1;
    if (monochrome)
	_our_mono_scheme = color_scheme;
    else
	_our_color_scheme = color_scheme;
}

